import numpy as np
from scipy.spatial import Delaunay
from hierarchicalDBSCAN.Robustness_tests import generate_dataset

np.random.seed(17041991)
colors = ['#C0223B', '#404ca0', 'rgba(173,216,230, 0.5)']  # colors for vertices, edges and 2-simplexes


def sq_norm(v):
    """ returns the squared norm of vector v """
    return np.linalg.norm(v)**2


def circumcircle(points,simplex):
    A = [points[simplex[k]] for k in range(3)]
    M = [[1.0]*4]
    M += [[sq_norm(A[k]), A[k][0], A[k][1], 1.0 ] for k in range(3)]
    M = np.asarray(M, dtype=np.float32)
    S = np.array([0.5*np.linalg.det(M[1:,[0,2,3]]), -0.5*np.linalg.det(M[1:,[0,1,3]])])
    a = np.linalg.det(M[1:, 1:])
    b = np.linalg.det(M[1:, [0,1,2]])
    return S/a,  np.sqrt(b/a+sq_norm(S)/a**2)


def get_alpha_complex(alpha, points, simplexes):
    """
    :param alpha: range from 0 to 1
    :param points: points which should be enclosed by the alpha complex
    :param simplexes: list of indices in the array of points that define 2-simplexes in the Delaunay triangulation
    :return:
    """
    return filter(lambda simplex: circumcircle(points,simplex)[1]<alpha, simplexes)


def get_area_of_alpha_complex(alpha_complex, dataset):
    """ computes the enclosed area of the alpha complex, by summing up the area of each triangle of the complex"""
    area = 0
    for s in alpha_complex:  # triangles of the alpha complex
        if len(s) == 3:
            A = dataset[s[0]]
            B = dataset[s[1]]
            C = dataset[s[2]]

            AB = np.linalg.norm(B-A)
            AC = np.linalg.norm(C-A)
            BC = np.linalg.norm(C-B)

            # calculate the semi-perimeter
            s = (AB + AC + BC) / 2

            # calculate the area
            area += ((s * (s - AB) * (s - AC) * (s - BC)) ** 0.5)

    return area


if __name__ == "__main__":
    dataset, labels_true = generate_dataset("moons", {"n_samples": 300, "noise": 0.075, "random_state": None})

    tri = Delaunay(dataset)
    alpha_complex_1 = get_alpha_complex(0.50, dataset, tri.simplices)
    area1 = get_area_of_alpha_complex(alpha_complex_1, dataset)

    pts = dataset[np.where(labels_true == 0), ][0]
    tri = Delaunay(pts)
    alpha_complex_1 = get_alpha_complex(0.50, pts, tri.simplices)
    area2 = get_area_of_alpha_complex(alpha_complex_1, pts)

    pts = dataset[np.where(labels_true == 1), ][0]
    tri = Delaunay(pts)
    alpha_complex_1 = get_alpha_complex(0.50, pts, tri.simplices)
    area3 = get_area_of_alpha_complex(alpha_complex_1, pts)

    print(area1, area2, area3)
