
from scipy.spatial import ConvexHull
from scipy.spatial.qhull import QhullError
from scipy.spatial.distance import euclidean
from shapely.geometry import Polygon
from shapely.geometry import Point
from shapely.prepared import prep
import numpy as np

__author__ = 'marco'




###########
# Mapping #
###########

# ps      ... list of points like array([c1, c2])
# point   ... numpy array([c1, c2]), c1 and c2 being coordinates.
# returns ... the index of point within list of points ps
#             or None, if point was not found in ps.
def find_point(ps, point, slacky=False):
    try:
        if slacky:
            return np.argwhere(np.all(np.isclose((ps-point), np.vstack((np.zeros(len(ps)), np.zeros(len(ps)))).T, atol=1e-8), axis=1))[0][0]
        else:
            return np.argwhere(np.all((ps-point) == 0, axis=1))[0][0]
    except IndexError:
        print(' - Index Error - ')
        return None


# returns ... a list of int [i1, i2, i3, ...,  i_n],
#             n being the length of new_ps
#             with i1, ... being the indices of point1, point2 ... from the new point list within the former point list
def map_point_set(sub_ps, former_ps):
    idx = [find_point(former_ps, sub_point) for sub_point in sub_ps]
    return idx


# edge_list ... like [(idx1, idx2), (idx2, idx3), ... ]
# returns   ... same list, but replaces idx_n with map[idx_n]
def map_edge_list(edge_list, my_map):
    return [(my_map[edge[0]], my_map[edge[1]]) for edge in edge_list]


# hull_edges ... an edge list like [[p1, p2], [p2, p3], ...], with p1, ... being int (indices in the orig. point set)
# returns    ... longest_edge, longest_length
#   longest_edge   ... longest edge like [p1, p2]
#   longest_length ... euclidean length of the longest edge
def get_longest_edge(ps, hull_edges, rank=0):
    hull_edges_with_length = []
    for simplex in hull_edges:
        hull_edges_with_length.append((simplex[0], simplex[1], euclidean(ps[simplex[0]], ps[simplex[1]])))
        # now hull_edges_with_length like [ (p1, p2, dist(p1,p2)), ... ]

    hull_edges_with_length.sort(key=lambda edge: edge[2], reverse=True)

    if len(hull_edges_with_length) > rank:
        longest_edge = (hull_edges_with_length[rank][:-1])
        longest_length = hull_edges_with_length[rank][2]
        return longest_edge, longest_length
    else:
        return None, None


###################
# Edge Conversion #
###################

# hull    ... vertices [a, b, c, ..., z], where a, b, ... are indices of points in the original point set.
# returns ... a list of edges (by indices), such as [[a, b], [b, c], ..., [z, a]]
def hull_to_edge_list(hull):
    fs = []
    for i in range(len(hull.vertices)):
        fs.append([hull.vertices[i], hull.vertices[(i+1) % len(hull.vertices)]])  # modulo: cycling last -> first elem.
    return fs


# edge_list ... a list of edges (by indices), such as [[a, b], [b, c], ..., [z, a]]
#               (with a, b being indices of points in the original point set)
# returns   ... list of vertices [a, b, c, ..., z]
def edge_list_to_hull(edge_list):
    return [edge[0] for edge in edge_list]


def invert_edge(edge):
    return (edge[1], edge[0])


def invert_edge_chain(edge_chain):
    return map(invert_edge, reversed(edge_chain))


#################
# Edge Rotation #
#################

# edges   ... list like [[p1, p2], [p2, p3], ...], with p1, ... being indices of points in the original point set
# idx     ... index where the split was done.
# returns ... the edges list in a rotated way, such that the last and first index is idx
def rotate_chain_to_index(edge_list, idx):
    # rotate, until split point index is at start of list
    while True:
        edge_list += [edge_list.pop(0)]
        if edge_list[-1][1] == idx:
            break
    return edge_list


def rotate_chain_to_edge(edge_list, edge, check_symmetrical_edges=False):
    # rotate, until split point index is at start of list
    while True:
        edge_list += [edge_list.pop(0)]
        if edge_list[-1] == edge or (check_symmetrical_edges and edge_list[-1] == invert_edge(edge)):
            break
    return edge_list


def follow_chain_until_index(edge_list, idx):
    sub_edge_list = []
    for edge in edge_list:
        sub_edge_list.append(edge)
        if edge[1] == idx:
            break
    return sub_edge_list


def follow_chain_until_edge(edge_list, edge, check_symmetrical_edges=False):
    sub_edge_list = []
    for my_edge in edge_list:
        sub_edge_list.append(my_edge)
        if my_edge == edge or (check_symmetrical_edges and my_edge == invert_edge(edge)):
            break
    return sub_edge_list


def subchain(edge_list, start_idx, end_idx):
    edge_list = rotate_chain_to_index(edge_list, start_idx)
    return follow_chain_until_index(edge_list, end_idx)


def subchain_by_edge(edge_list, start_edge, end_edge, check_symmetrical_edges=False):
    edge_list = rotate_chain_to_edge(edge_list, start_edge, check_symmetrical_edges)
    return follow_chain_until_edge(edge_list, end_edge, check_symmetrical_edges)



########
# Hull #
########

# ps      ... a point set
# returns ... an edge list such as [[p1, p2], [p2, p3], ...] with p1, ... being int (indices in the original point set)
def get_convex_hull(ps):
    if len(ps) < 3:
        return None

    try:
        hull = ConvexHull(ps)
    except QhullError:
        print(" - Qhull Error - ")
        return None

    hull_edges = []
    # facets(hull) like [[p1, p2], [p2, p3], ...] with p1, p2, p3, ... being int (indices in the original point set ps)
    for simplex in hull_to_edge_list(hull):
        # simplex like [p1, p2], with p1 and p2 being indices of points in the original point set ps.
        hull_edges.append((simplex[0], simplex[1]))

    return hull_edges


# ps         ... a point set
# edge_chain ... list of edges like [(idx1, idx2), ...] with idx1, ... being indices within the point set
# returns    ... all points of ps that lie within the borders of edge_chain
def find_points_in_hull(ps, edge_chain):
    polygon = Polygon(ps[edge_list_to_hull(edge_chain)])
    polygon = polygon.buffer(0.0000000001)
    points = [Point(p[0], p[1]) for p in ps]
    prepared_polygon = prep(polygon)
    ps_hits = filter(prepared_polygon.contains, points)
    ps_hits = np.asarray([(p.x, p.y) for p in ps_hits])
    return ps_hits


