from sklearn.metrics import *
import numpy as np
from itertools import compress
import math
import scipy.stats as stats
import matplotlib.pyplot as plt
from .ModifiedSilhouetteCoefficient import modified_silhouette_score
from hierarchicalDBSCAN.RandSelect import *
import networkx


def filter_noise(dataset, labels):
    """ removes points labeled as noise from the dataset

    :param dataset: dataset to be filtered
    :param labels: labels including noise classifications (-1)
    :return: return filtered dataset and filtered labellist
    """
    filter = np.array([l >= 0 for l in labels])
    dataset = dataset[np.array(filter)]
    labels = list(compress(labels, filter))
    return dataset, labels



def get_core_distance(distance_list, minPts, n):
    """ Returns the distance of the k-th nearest neighbor

    :param "distance_list": pairwise distances of the point to all points
    :param "minPts" : minimum number of points of the neighborhood set
    :param "n": number of points in the dataset

    :reutnr: binning distance matrix into equal sized bins in the range of (0, max_range)
    """
    return RandSelect(distance_list, minPts, n)


def get_modified_silhouette_score(dist_mat, labels, minPts):
    """ calculate density based silhouette score for one labeling

    :param dist_mat: distance matrix of a dataset
    :param labels: labels of the clustering
    :param minPts: used minPts necessary for calculating core distance
    :return: density based silhouette score
    """
    N = len(dist_mat)
    core_distance = [get_core_distance(dist_mat[x, :], minPts, N) for x in range(N)]

    clusterheight = {l: 0 for l in set(labels)}

    graph = networkx.empty_graph(N)
    for source in range(N):
        for target in range(source+1, N):
            dist = max(min(core_distance[source], core_distance[target]), dist_mat[source, target])
            if np.inf >= dist:
                graph.add_edge(source, target, weight=dist)

    for label in set(labels):
        indexlist = np.where(np.array(labels) == label)[0].tolist()
        subgraph = graph.subgraph(indexlist)
        clustergraph = networkx.minimum_spanning_tree(subgraph)
        edge_weights = [clustergraph.edge[u][v]["weight"] for (u, v) in clustergraph.edges()]
        if len(indexlist) == 1:
            edge_weights += [core_distance[indexlist[0]]]
        clusterheight[label] = max(edge_weights)

    return modified_silhouette_score(dist_mat, np.array(labels), clusterheight)


def optimize_modified_silhouette(dataset, label_gen, dist_mat, dont_filter=True, k=None, fixed_height=None):
    """ calculate density based silhouette score for each labeling of a labelgenerator

    :param dataset: used dataset
    :param label_gen: labelgenerator
    :param dist_mat: distance matrix of the datset
    :param dont_filter: filter noise points yes or no
    :param k: only rate labelings with exactly k clusters
    :param fixed_height: use fixed eps instead of clusterheights
    :return: best hierarchy level and its score
    """
    N = len(dataset)
    best = {"height": 0, "score": -1}

    for (height, labels, clusterheights) in label_gen:
        if fixed_height is not None:
            clusterheights = {x: fixed_height for x in clusterheights}
        if dont_filter:
            filtered_dataset, filtered_labels = (dataset, labels)
        else:
            filtered_dataset, filtered_labels = filter_noise(dataset, labels)
        if len(set(filtered_labels)) < 2 or len(set(filtered_labels)) == N:
            continue
        if k is not None and len(set(filtered_labels)) != k:
            continue

        score = modified_silhouette_score(dist_mat, np.array(filtered_labels), clusterheights, metric="precomputed")

        if score > best["score"]:
            best["height"] = height
            best["score"] = score

    return best


def minimal_cluster_distance(distance_matrix, labels, i, j):
    """Calculate the minimal distance between cluster i and j

    Parameters
    ----------

    i : int
        Label of the first cluster
    j : int
        Label of the second cluster

    Returns
    -------
    b : float
        Mean nearest-cluster distance for sample i
    """
    indices_c1 = np.where(np.array(labels) == i)
    indices_c2 = np.where(np.array(labels) == j)
    return np.min(distance_matrix[indices_c1,:][0][:, indices_c2])


def optimize_silhouette(dataset, label_gen, dist_mat, k=None, dont_filter=True):
    """ calculate standard silhouette score for each labeling of a labelgenerator

    :param dataset: used dataset
    :param label_gen: labelgenerator
    :param dist_mat: distance matrix of the datset
    :param dont_filter: filter noise points yes or no
    :param k: only rate labelings with exactly k clusters
    :return: best hierarchy level and its score
    """
    N = len(dataset)
    best = {"height": 0, "score": -1}

    for height, labels, t in label_gen:
        if dont_filter:
            filtered_dataset, filtered_labels = (dataset, labels)
        else:
            filtered_dataset, filtered_labels = filter_noise(dataset, labels)
        if len(set(filtered_labels)) < 2 or len(set(filtered_labels)) == N:
            continue
        if k is not None and len(set(filtered_labels)) != k:
            continue

        score = silhouette_score(dist_mat, np.array(filtered_labels), metric="precomputed")
        if score > best["score"]:
            best["height"] = height
            best["score"] = score

    return best


def optimize_edge_correlation(dataset, label_gen, k=None, dist_mat=None, plot=False):
    """ calculate edge correlation score for each labeling of a labelgenerator

    :param dataset: used dataset
    :param label_gen: labelgenerator
    :param dist_mat: distance matrix of the datset
    :param k: only rate labelings with exactly k clusters
    :param plot: plot edge correlation score of each level
    :return: best hierarchy level and its score
    """
    best = {"height": 0, "score": -1}
    N = len(dataset)

    if dist_mat is None:
        distance_matrix = pairwise_distances(dataset, metric="euclidean")
    else:
        distance_matrix = dist_mat

    similarity = 1 - distance_matrix
    scores = []

    for (height, labels, _) in label_gen:
        #todo include filter of similarity matrix and label vector
        filtered_dataset, filtered_labels = (dataset, labels)

        if len(set(filtered_labels)) < 2 or len(set(filtered_labels)) == N:
            continue
        if k is not None and len(set(filtered_labels).difference(set([-1, -2]))) != k:
            continue

        labelmatrix = pairwise_distances(np.transpose(np.matrix(filtered_labels)), metric="hamming")

        score, _ = stats.pearsonr(similarity.flatten(), labelmatrix.flatten())

        score = math.fabs(score)
        if plot:
            scores.append((height, score))

        if score > best["score"]:
            best["height"] = height
            best["score"] = score

    if plot:
        [heights, scores] = zip(*scores)
        plt.scatter(heights, scores)
        plt.show()

    return best
