import numpy as np
from sklearn.utils import check_random_state
from sklearn.metrics.pairwise import pairwise_distances
from sklearn.metrics.cluster.unsupervised import _nearest_cluster_distance

def modified_silhouette_score(X, labels, clusterheights, metric="euclidean", sample_size=None,
                     random_state=None, **kwds):
    """Compute the mean density based Silhouette Coefficient of all samples.

    The density based Silhouette Coefficient is calculated using the cluster
    generation distance (``a``) and the mean nearest-cluster distance (``b``) for each
    sample.  The Silhouette Coefficient for a sample is ``(b - a) / max(a,
    b)``.  To clarify, ``b`` is the distance between a sample and the nearest
    cluster that the sample is not a part of.
    Note that Silhouette Coefficent is only defined if number of labels
    is 2 <= n_labels <= n_samples - 1.

    This function returns the mean density based Silhouette Coefficient over all samples.
    To obtain the values for each sample, use :func:`silhouette_samples`.


    Parameters
    ----------
    X : array [n_samples_a, n_samples_a] if metric == "precomputed", or, \
             [n_samples_a, n_features] otherwise
        Array of pairwise distances between samples, or a feature array.

    labels : array, shape = [n_samples]
         Predicted labels for each sample.

    metric : string, or callable
        The metric to use when calculating distance between instances in a
        feature array. If metric is a string, it must be one of the options
        allowed by :func:`metrics.pairwise.pairwise_distances
        <sklearn.metrics.pairwise.pairwise_distances>`. If X is the distance
        array itself, use ``metric="precomputed"``.

    sample_size : int or None
        The size of the sample to use when computing the Silhouette
        Coefficient. If ``sample_size is None``, no sampling is used.

    random_state : integer or numpy.RandomState, optional
        The generator used to initialize the centers. If an integer is
        given, it fixes the seed. Defaults to the global numpy random
        number generator.

    `**kwds` : optional keyword parameters
        Any further parameters are passed directly to the distance function.
        If using a scipy.spatial.distance metric, the parameters are still
        metric dependent. See the scipy docs for usage examples.

    Returns
    -------
    silhouette : float
        Mean density based Silhouette Coefficient for all samples.

    References
    ----------

    .. [1] `Peter J. Rousseeuw (1987). "Silhouettes: a Graphical Aid to the
       Interpretation and Validation of Cluster Analysis". Computational
       and Applied Mathematics 20: 53-65.
       <http://www.sciencedirect.com/science/article/pii/0377042787901257>`_

    .. [2] `Wikipedia entry on the Silhouette Coefficient
           <http://en.wikipedia.org/wiki/Silhouette_(clustering)>`_

    .. [3] 'Alexander Dockhorn, Christian Braune, Rudolf Kruse (2015). An
        Alternating Optimization Approach based on Hierarchical Adaptations
        of DBSCAN. handed in to SSCI 2015

    """
    n_labels = len(np.unique(labels))
    n_samples = X.shape[0]
    if not 1 < n_labels < n_samples:
        return 0
        """raise ValueError("Number of labels is %d. Valid values are 2 "
                         "to n_samples - 1 (inclusive)" % n_labels)"""

    if sample_size is not None:
        random_state = check_random_state(random_state)
        indices = random_state.permutation(X.shape[0])[:sample_size]
        if metric == "precomputed":
            X, labels = X[indices].T[indices].T, labels[indices]
        else:
            X, labels = X[indices], labels[indices]
    #print("clusterheights", clusterheights)
    return np.mean(modified_silhouette_samples(X, labels, clusterheights, metric=metric, **kwds))


def modified_silhouette_samples(X, labels, clusterheights, metric='euclidean', **kwds):
    """Compute the density based Silhouette Coefficient for each sample.

    The Silhouette Coefficient is a measure of how well samples are clustered
    with samples that are similar to themselves. Clustering models with a high
    Silhouette Coefficient are said to be dense, where samples in the same
    cluster are similar to each other, and well separated, where samples in
    different clusters are not very similar to each other.

    The density based Silhouette Coefficient is calculated using the cluster
    generation distance (``a``) and the mean nearest-cluster distance (``b``) for each
    sample.  The Silhouette Coefficient for a sample is ``(b - a) / max(a,
    b)``.
    Note that Silhouette Coefficent is only defined if number of labels
    is 2 <= n_labels <= n_samples - 1.

    This function returns the Silhouette Coefficient for each sample.


    Parameters
    ----------
    X : array [n_samples_a, n_samples_a] if metric == "precomputed", or, \
             [n_samples_a, n_features] otherwise
        Array of pairwise distances between samples, or a feature array.

    labels : array, shape = [n_samples]
             label values for each sample

    metric : string, or callable
        The metric to use when calculating distance between instances in a
        feature array. If metric is a string, it must be one of the options
        allowed by :func:`sklearn.metrics.pairwise.pairwise_distances`. If X is
        the distance array itself, use "precomputed" as the metric.

    `**kwds` : optional keyword parameters
        Any further parameters are passed directly to the distance function.
        If using a ``scipy.spatial.distance`` metric, the parameters are still
        metric dependent. See the scipy docs for usage examples.

    Returns
    -------
    silhouette : array, shape = [n_samples]
        density based Silhouette Coefficient for each samples.

    References
    ----------

    .. [1] `Peter J. Rousseeuw (1987). "Silhouettes: a Graphical Aid to the
       Interpretation and Validation of Cluster Analysis". Computational
       and Applied Mathematics 20: 53-65.
       <http://www.sciencedirect.com/science/article/pii/0377042787901257>`_

    .. [2] `Wikipedia entry on the Silhouette Coefficient
       <http://en.wikipedia.org/wiki/Silhouette_(clustering)>`_

    .. [3] 'Alexander Dockhorn, Christian Braune, Rudolf Kruse (2015). An
        Alternating Optimization Approach based on Hierarchical Adaptations
        of DBSCAN. handed in to SSCI 2015

    """
    distances = pairwise_distances(X, metric=metric, **kwds)
    n = labels.shape[0]
    clusterheights[-1] = np.inf
    clusterheights[-2] = 0
    A = np.array([clusterheights[l] for l in labels])
    B = np.array([_nearest_cluster_distance(distances[i], labels, i)
                  for i in range(n)])
    #min_dist = {i: min([minimal_cluster_distance(distances,labels,i,j) for j in set(labels)]) for i in set(labels)}
    #B = np.array([min_dist[labels[i]] for i in range(n)])
    sil_samples = (B - A) / np.maximum(A, B)
    # nan values are for clusters of size 1, and should be 0
    return np.nan_to_num(sil_samples)


def minimal_cluster_distance(distance_matrix, labels, i, j):
    """Calculate the minimal distance between cluster i and j

    Parameters
    ----------

    i : int
        Label of the first cluster
    j : int
        Label of the second cluster

    Returns
    -------
    b : float
        Mean nearest-cluster distance for sample i
    """
    indices_c1 = np.where(np.array(labels) == i)
    indices_c2 = np.where(np.array(labels) == j)
    return np.min(distance_matrix[indices_c1,:][0][:, indices_c2])