__author__ = 'Alex'

import numpy as np
from hierarchicalDBSCAN.Datasets.DatasetLoader import *
from sklearn.metrics.cluster.supervised import homogeneity_completeness_v_measure
import networkx as nx
from hierarchicalDBSCAN.ClusterValidation.InternalValidation import get_modified_silhouette_score
from sklearn.metrics.pairwise import pairwise_distances

class TwoDimClique():
    """ Clique implementation for two dimensional datasets. Datasets with higher dimensionality will only be clustered
    based on their first two dimensions.
    """

    def __init__(self, gridsizes=[(10, 10), (15, 15), (20, 20), (25, 25)], threshold=3):
        """ Set up gridsizes and threshold to use for clustering

        :param gridsizes: list of tuples of gridsizes to use
        :param threshold: threshold for every gridcell to be dense
        """
        self.gridsizes = gridsizes
        self.threshold = threshold
        self.labels = None

    def fit(self, dataset):
        """ fit a clustering for the dataset

        :param dataset: dataset to be clusterd
        """
        best_score = -1
        best_pred  = [-1]*len(dataset)

        dist_mat = pairwise_distances(dataset)

        for gridsize in self.gridsizes:
            min_x = min(dataset[:, 0])
            max_x = max(dataset[:, 0])
            min_y = min(dataset[:, 1])
            max_y = max(dataset[:, 1])
            x_stepsize = (max_x-min_x)/gridsize[0]
            y_stepsize = (max_y-min_y)/gridsize[1]

            counting_graph = nx.grid_2d_graph(gridsize[0]+1, gridsize[1]+1)
            nx.set_node_attributes(counting_graph, 'points', {n: [] for n in counting_graph.node})
            for i in range(dataset.shape[0]):
                x = int((dataset[i, 0] - min_x) // x_stepsize)
                y = int((dataset[i, 1] - min_y) // y_stepsize)
                counting_graph.node[(x, y)]["points"].append(i)
            pred = self.find_cluster(counting_graph, dataset)
            score = get_modified_silhouette_score(dist_mat, pred, self.threshold)

            if score > best_score:
                best_pred = pred
                best_score = score
        self.labels = best_pred


    def find_cluster(self, counting_graph, dataset):
        """ merge neighboring dense gridcells to one cluster

        :param counting_graph: networkx graph for the gridded dataspace
        :param dataset: dataset to be clustered
        :return: clustering of the dataset
        """
        del_edges = [(u, v) for (u, v) in counting_graph.edges()
                     if len(counting_graph.node[u]["points"]) < self.threshold or
                     len(counting_graph.node[v]["points"]) < self.threshold]
        counting_graph.remove_edges_from(del_edges)

        labels = {point: -1 for point in range(len(dataset))}
        for c_idx, subgraph in enumerate(nx.connected_component_subgraphs(counting_graph)):
            if len(subgraph) < self.threshold:
                for node in subgraph:
                    for p_idx in counting_graph.node[node]["points"]:
                        labels[p_idx] = -1
            else:
                for node in subgraph:
                    for p_idx in counting_graph.node[node]["points"]:
                        labels[p_idx] = c_idx
        return [labels[point] for point in range(len(dataset))]


    def predict(self, dataset):
        """ return clustering of the fitted dataset

        :param dataset: dataset to be clustered in case fit() was not runned yet
        :return: clustering of the dataset
        """
        if self.labels is None:
            self.fit(dataset)
        return self.labels


if __name__ == "__main__":
    dataset, labels_true = standard()

    clique = TwoDimClique(threshold=4)
    clique.fit(dataset)

    labels_pred = clique.predict(dataset)

    print("q = ", homogeneity_completeness_v_measure(labels_true, labels_pred))
