__author__ = 'Alex'

from hierarchicalDBSCAN.ClusteringAlgorithms.minPtsHierarchicalDBScan import *
from hierarchicalDBSCAN.Datasets.DatasetLoader import *
from sklearn.metrics.cluster.supervised import homogeneity_completeness_v_measure
import numpy as np
import prettyplotlib as ppl
from hierarchicalDBSCAN.ClusterValidation.InternalValidation import get_modified_silhouette_score
import hierarchicalDBSCAN.AlphaShapes.Alpha2D as alpha
from scipy.spatial import Delaunay
#import pygraphviz

class DifDenseAlphaDBSCAN:
    """ HDBSCAN based on Clustertree implementing edge quantile cut
    """
    def __init__(self, minPts):
        self.minPts = minPts
        self.db = minPtsHierarchicalDBScan(self.minPts)
        self.dataset = None


    def fit(self, dataset):
        """ run minPtsHierarchicalDBSCAN on the provided dataset

        :param dataset: dataset to be clustered
        """
        self.dataset = dataset
        self.db.fit(dataset)
        self.db.simplify_clustergraph()


    def predict(self, dataset, nr_clusters = None):
        """ tries different alpha cuts and reports the cut with the best density-based silhouette score

        :param dataset: dataset to be clustered in case fit was not used before
        :return: labeling of the best alpha cut
        """
        if self.dataset is None:
            self.fit(dataset)
        if nr_clusters is None:
            nr_clusters = len(dataset)*0.02

        hierarchy_nodes = []
        densitychanges = list()
        changes = []

        for (s, t) in self.db.clustergraph.edges():
            pts_parent = dataset[list(self.db.clustergraph.node[s]['nodes']), :]
            triparent = Delaunay(pts_parent)
            alphacomplexparent = alpha.get_alpha_complex(0.50, pts_parent, triparent.simplices)
            areaparent = alpha.get_area_of_alpha_complex(alphacomplexparent, pts_parent) + 0.1
            densityparent = len(pts_parent) / areaparent

            pts_child = dataset[list(self.db.clustergraph.node[t]['nodes']), :]
            trichild = Delaunay(pts_child)
            alphacomplexchild = alpha.get_alpha_complex(0.50, pts_child, trichild.simplices)
            areachild = alpha.get_area_of_alpha_complex(alphacomplexchild, pts_child) + 0.1
            densitychild =  len(pts_child) / areachild

            densitychanges.append((areaparent - areachild, self.db.clustergraph.node[t]["nodes"], self.db.clustergraph.node[t]["height"]))
            changes.append(areaparent/areachild)

        a = sorted(densitychanges, key=itemgetter(0), reverse=True)

        b = a[0:nr_clusters]
        c = sorted(b, key=itemgetter(2))

        # print("heights", [y for _,y in hierarchy_nodes])
        default = 1
        labels = [default] * len(dataset)
        for c_idx in range(len(c)):
            for p_idx in c[c_idx][1]:
                if labels[p_idx] == default:
                    labels[p_idx] = c_idx+2
        return labels




if __name__ == "__main__":
    #dataset, labels_true = spiral()
    #dataset, labels_true = compound()
    #dataset, labels_true = aggregation()
    #dataset, labels_true = t48k(size=2000)
    dataset, labels_true = flame()
    #dataset, labels_true = standard()
    #dataset, labels_true = jain()
    #dataset, labels_true = r15()
    #dataset, labels_true = generate_dataset("moons", {"n_samples":300, "noise": 0.075, "random_state":None})
    #dataset, labels_true = r15()

    nr_clusters = 1
    difdense = DifDenseAlphaDBSCAN(5)
    difdense.fit(dataset)

    labels_pred = difdense.predict(dataset, nr_clusters)
    print(labels_pred)


    print("q = ", homogeneity_completeness_v_measure(labels_true, labels_pred))
    color = [str(item/255.) for item in labels_pred]


    base_colors = [(255,187,120),(255,127,14),(174,199,232),(44,160,44),(31,119,180),(255,152,150),(214,39,40),
               (197,176,213),(152,223,138),(148,103,189),(247,182,210),(227,119,194),(196,156,148),(140,86,75),
               (127,127,127),(219,219,141),(199,199,199),(188,189,34),(158,218,229),(23,190,207)]*10

    colors = {label: [t/255 for t in base_colors[i]] for i, label in enumerate(set(labels_pred))}

    labeling = labels_pred
    data = dataset
    fig = plt.figure()
    labeling = np.asarray(labeling)
    if len(set(labeling)) > 1:
        colors = {label: [t/255 for t in base_colors[i]] for i, label in enumerate(set(labeling))}
        for label in set(labeling):
            filtered_dataset = data[np.where(labeling==label)[0].tolist(), ]

            #filtered_dataset = dataset[np.where(labeling == label)[0].tolist(),:]
            ppl.scatter(plt.gca(), filtered_dataset[:, 0], filtered_dataset[:, 1], label=str(label),
                        color=colors[label], linewidth=1, edgecolor='black')
    else:
        ppl.scatter(plt.gca(), data[:, 0], data[:, 1],
                    color='black', linewidth=1, edgecolor='black', alpha=1)
    plt.show()