__author__ = 'Alex'

from hierarchicalDBSCAN.ClusteringAlgorithms.minPtsHierarchicalDBScan import *
from hierarchicalDBSCAN.Datasets.DatasetLoader import *
from sklearn.metrics.cluster.supervised import homogeneity_completeness_v_measure
import numpy as np
import prettyplotlib as ppl
from hierarchicalDBSCAN.ClusterValidation.InternalValidation import get_modified_silhouette_score

class DifDenseDBSCAN:
    """ HDBSCAN based on Clustertree implementing edge quantile cut
    """
    def __init__(self, minPts):
        self.minPts = minPts
        self.db = minPtsHierarchicalDBScan(self.minPts)
        self.dataset = None


    def fit(self, dataset):
        """ run minPtsHierarchicalDBSCAN on the provided dataset

        :param dataset: dataset to be clustered
        """
        self.dataset = dataset
        self.db.fit(dataset)
        self.db.simplify_clustergraph()


    def predict(self, dataset):
        """ tries different alpha cuts and reports the cut with the best density-based silhouette score

        :param dataset: dataset to be clustered in case fit was not used before
        :return: labeling of the best alpha cut
        """
        if self.dataset is None:
            self.fit(dataset)

        dendrogram_edges = [self.db.clustergraph.node[u]["height"] - self.db.clustergraph.node[v]["height"]
                            for (u, v) in self.db.simplified_clustergraph.edges()]
        # convert to standard gaussian distribution with mean = 0 and sd = 1

        dist_mat = pairwise_distances(dataset)

        best_score = -1
        best_pred  = [-1]*len(dataset)

        alphas = [1.28, 1.44, 1.654, 1.96, 2.33, 2.58]
        for alpha in alphas:
            height = alpha*math.sqrt(np.var(dendrogram_edges))+np.mean(dendrogram_edges)
            pred = self.db.predict_cut_edge_quantile(self.db.simplified_clustergraph, height)
            if len(set(pred)) > 1:
                score = get_modified_silhouette_score(dist_mat, pred, self.minPts)
                if score > best_score:
                    best_pred = pred
                    best_score = score

        return best_pred


if __name__ == "__main__":
    #dataset, labels_true = spiral()
    #dataset, labels_true = compound()
    #dataset, labels_true = aggregation()
    #dataset, labels_true = t48k(size=2000)
    #dataset, labels_true = flame()
    #dataset, labels_true = standard()
    #dataset, labels_true = r15()

    dataset, labels_true = standard()

    difdense = DifDenseDBSCAN(4)
    difdense.fit(dataset)

    labels_pred = difdense.predict(dataset)

    print("q = ", homogeneity_completeness_v_measure(labels_true, labels_pred))
    color = [str(item/255.) for item in labels_pred]


    base_colors = [(255,187,120),(255,127,14),(174,199,232),(44,160,44),(31,119,180),(255,152,150),(214,39,40),
               (197,176,213),(152,223,138),(148,103,189),(247,182,210),(227,119,194),(196,156,148),(140,86,75),
               (127,127,127),(219,219,141),(199,199,199),(188,189,34),(158,218,229),(23,190,207)]*10

    colors = {label: [t/255 for t in base_colors[i]] for i, label in enumerate(set(labels_pred))}

    labeling = labels_pred
    data = dataset
    fig = plt.figure()
    labeling = np.asarray(labeling)
    if len(set(labeling)) > 1:
        colors = {label: [t/255 for t in base_colors[i]] for i, label in enumerate(set(labeling))}
        for label in set(labeling):
            filtered_dataset = data[np.where(labeling==label)[0].tolist(), ]

            #filtered_dataset = dataset[np.where(labeling == label)[0].tolist(),:]
            ppl.scatter(plt.gca(), filtered_dataset[:, 0], filtered_dataset[:, 1], label=str(label),
                        color=colors[label], linewidth=1, edgecolor='black')
    else:
        ppl.scatter(plt.gca(), data[:, 0], data[:, 1],
                    color='black', linewidth=1, edgecolor='black', alpha=1)
    plt.show()