__author__ = 'Alex'
import logging
import numpy as np
from hierarchicalDBSCAN.ClusteringAlgorithms.epsDBScan import epsHierarchicalDBScan
from hierarchicalDBSCAN.ClusteringAlgorithms.minPtsHierarchicalDBScan import minPtsHierarchicalDBScan
from hierarchicalDBSCAN.Datasets.DatasetLoader import *
from hierarchicalDBSCAN.ClusterValidation.InternalValidation import *
from hierarchicalDBSCAN.Robustness_tests import generate_dataset
from sklearn.metrics import homogeneity_completeness_v_measure
from sklearn.cluster import DBSCAN
import colorsys
from hierarchicalDBSCAN.ClusteringAlgorithms.minPtsGraphDBScan import *

import time

DEFAULT_MINPTS = 5
DEFAULT_EPS = 0.3

class aoDBSCAN():
    """ alternating optimization implementation for DBSCAN
    """
    def __init__(self, start_eps=None, start_minpts=None, opt="modsil"):
        """ initialize cluster objects for both hierarhicl variants
        :param start_eps:
        :param start_minpts:
        :param opt:
        :return:
        """
        if start_eps is None:
            self.db_eps = epsHierarchicalDBScan(eps=DEFAULT_EPS)
        else:
            self.db_eps = epsHierarchicalDBScan(eps=start_eps)

        if start_minpts is None:
            #self.db_minPts = minPtsHierarchicalDBScan(minPts=DEFAULT_MINPTS)
            self.db_minPts = minPtsGraphDBScan(minPts=DEFAULT_MINPTS)
        else:
            #self.db_minPts = minPtsHierarchicalDBScan(minPts=start_minpts)
            self.db_minPts = minPtsGraphDBScan(minPts=start_minpts)

        self.opt=opt
        self.dataset = None
        self.current_optimum = None

    def fit(self, dataset):
        """ default fitting method = fit_minPts

        :param dataset: dataset to be clustered
        """
        self.fit_minPts(dataset, start_minPts=5)

    def fit_eps(self, dataset, start_eps=None, iterations=10, bins=None, plot_iterations=False):
        """ start alternating optimization using fit_eps

        :param dataset: dataset to be clustered
        :param start_eps: starting epsilon value
        :param iterations: maximal number of iterations
        :param bins: binning of the distance matrix
        :param plot_iterations: plot all iterations of the ao process
        """
        if start_eps is None:
            start_eps = self.db_eps
        best_eps = start_eps
        best_minPts = -1
        N = len(dataset)
        self.dataset = dataset
        self.current_optimum = {"score": -1, "radius": best_eps, "minPts": best_minPts}
        processed_minPts = []
        processed_eps = []
        parameter_combinations = []

        dist_mat = pairwise_distances(dataset)

        for i in range(iterations):
            print("iteration", i)
            #check for loop
            if best_eps in processed_eps:
                logging.debug("optimization does loop")
                break
            processed_eps.append(best_eps)

            #prepare and fit dbscan
            self.db_eps.configure(eps=best_eps)
            self.db_eps.fit(dataset, dist_mat=dist_mat)
            self.db_eps.simplify_clustergraph()

            #determine optimum
            if self.opt=="modsil":
                best = optimize_modified_silhouette(dataset, self.db_eps.get_clusterings(), dist_mat, fixed_height=best_eps)
            elif self.opt=="sil":
                best = optimize_silhouette(dataset, self.db_eps.get_clusterings(), dist_mat)
            elif self.opt=="edge":
               best = optimize_edge_correlation(dataset, self.db_eps.get_clusterings(), dist_mat=dist_mat, plot=False)
            else:
                print("choose modsil, sil or edge for optimization")
                return

            best_minPts = N-best["height"]+1
            if best["score"] > self.current_optimum["score"]:
                self.current_optimum["score"] = best["score"]
                self.current_optimum["radius"] = best_eps
                self.current_optimum["minPts"] = best_minPts
                self.current_optimum["height"] = best["height"]
                self.current_optimum["dbscan"] = "eps"
                parameter_combinations.append((best_eps, best_minPts))
            if plot_iterations:
                self.db_eps.plot(True, best["height"])


            #check for loop
            if best_minPts in processed_minPts:
                logging.debug("optimization does loop")
                break
            processed_minPts.append(best_minPts)

            #prepare and fit dbscan
            self.db_minPts.configure(minPts=best_minPts)
            self.db_minPts.fit(dataset, dist_mat=dist_mat, bins=bins)
            self.db_minPts.simplify_clustergraph()

            #determine optimum
            if self.opt=="modsil":
                best = optimize_modified_silhouette(dataset, self.db_minPts.get_clusterings(), dist_mat)
            elif self.opt == "sil":
                best = optimize_silhouette(dataset, self.db_minPts.get_clusterings(), dist_mat)
            elif self.opt=="edge":
                best = optimize_edge_correlation(dataset, self.db_minPts.get_clusterings(), dist_mat=dist_mat, plot=False)
            else:
                print("choose modsil, sil or edge for optimization")
                return

            best_eps = best["height"]
            if best["score"] > self.current_optimum["score"]:
                self.current_optimum["score"] = best["score"]
                self.current_optimum["height"] = best_eps
                self.current_optimum["radius"] = best_eps
                self.current_optimum["minPts"] = best_minPts
                self.current_optimum["dbscan"] = "minPts"
                parameter_combinations.append((best_eps, best_minPts))
            if plot_iterations:
                self.db_minPts.plot(True, best["height"])
        print("found optimum", self.current_optimum, "after", i+1, "iterations")
        self.params = parameter_combinations
        return parameter_combinations


    def fit_minPts(self, dataset, start_minPts=None, iterations=10, bins=None, plot_iterations=False, optimization="modsil"):
        """ start alternating optimization using fit_minPts

        :param dataset: dataset to be clustered
        :param start_eps: starting epsilon value
        :param iterations: maximal number of iterations
        :param bins: binning of the distance matrix
        :param plot_iterations: plot all iterations of the ao process
        """
        if start_minPts is None:
            start_minPts = self.db_minPts
        best_eps = -1
        best_minPts = start_minPts
        N = len(dataset)
        self.dataset = dataset
        self.current_optimum = {"score": -1, "radius": best_eps, "minPts": best_minPts}
        processed_minPts = []
        processed_eps = []

        dist_mat = pairwise_distances(dataset)
        parameter_combinations = []
        for i in range(iterations):
            logging.debug("iteration {}".format(i))
            #check for loop
            if best_minPts in processed_minPts:
                logging.debug("optimization does loop")
                break
            processed_minPts.append(best_minPts)

            logging.debug("fit minPtsdbs")
            #prepare and fit dbscan
            self.db_minPts.configure(minPts=best_minPts)
            self.db_minPts.fit(dataset, dist_mat=dist_mat, bins=bins)
            self.db_minPts.simplify_clustergraph()
            logging.debug("done")

            #determine optimum
            if optimization=="modsil":
                best = optimize_modified_silhouette(dataset, self.db_minPts.get_clusterings(), dist_mat)
            elif optimization=="sil":
                best = optimize_silhouette(dataset, self.db_minPts.get_clusterings(), dist_mat)
            elif optimization=="edge":
                best = optimize_edge_correlation(dataset, self.db_minPts.get_clusterings(), dist_mat=dist_mat, plot=False)
            else:
                print("choose modsil, sil or edge for optimization")
                return

            best_eps = best["height"]
            parameter_combinations.append((best_eps, best_minPts))

            if best["score"] > self.current_optimum["score"]:
                self.current_optimum["score"] = best["score"]
                self.current_optimum["height"] = best_eps
                self.current_optimum["radius"] = best_eps
                self.current_optimum["minPts"] = best_minPts
                self.current_optimum["dbscan"] = "minPts"
            if plot_iterations:
                self.db_minPts.plot(True, best["height"])


            #check for loop
            if best_eps in processed_eps:
                logging.debug("optimization does loop")
                break
            processed_eps.append(best_eps)

            logging.debug("fit epsdbs")
            #prepare and fit dbscan
            self.db_eps.configure(eps=best_eps)
            self.db_eps.fit(dataset, dist_mat=dist_mat)
            self.db_eps.simplify_clustergraph()
            logging.debug("done")

            #determine optimum
            if self.opt=="modsil":
                best = optimize_modified_silhouette(dataset, self.db_eps.get_clusterings(), dist_mat, fixed_height=best_eps)
            elif self.opt=="sil":
                best = optimize_silhouette(dataset, self.db_eps.get_clusterings(), dist_mat)
            elif self.opt=="edge":
                best = optimize_edge_correlation(dataset, self.db_eps.get_clusterings(), dist_mat=dist_mat, plot=False)
            else:
                print("choose modsil, sil or edge for optimization")
                return
            best_minPts = N-best["height"]+1
            parameter_combinations.append((best_eps, best_minPts))

            if best["score"] > self.current_optimum["score"]:
                self.current_optimum["score"] = best["score"]
                self.current_optimum["radius"] = best_eps
                self.current_optimum["minPts"] = best_minPts
                self.current_optimum["height"] = best["height"]
                self.current_optimum["dbscan"] = "eps"
            if plot_iterations:
                self.db_eps.plot(True, best["height"])
        logging.debug("found optimum {} after {} iterations".format(self.current_optimum, i+1))
        self.params = parameter_combinations
        return parameter_combinations

    def plot(self):
        """ plot found optimum
        """
        try:
            if self.dataset is None:
                raise ProcessingSequenceError('first run fit method')
        except ProcessingSequenceError:
            return

        if self.current_optimum["dbscan"] == "minPts":
            self.db_minPts.configure(minPts=self.current_optimum["minPts"])
            self.db_minPts.fit(self.dataset)
            self.db_minPts.plot(True, self.current_optimum["height"])
            return self.db_minPts.labels
        else:
            self.db_eps.configure(eps=self.current_optimum["radius"])
            self.db_eps.fit(self.dataset)
            self.db_eps.plot(True, self.current_optimum["height"])
            return self.db_minPts.labels

    def plot_learning(self):
        """ plot learning process
        """
        xs = np.array([x for x,_ in self.params])
        ys = np.array([y for _,y in self.params])
        print("params", params)
        plt.quiver(xs[:-1], ys[:-1], xs[1:]-xs[:-1], ys[1:]-ys[:-1], scale_units='xy', angles='xy', scale=1, width=0.01)
        plt.xlim(0.9*min(xs), 1.1*max(xs))
        plt.ylim(0.9*min(ys), 1.1*max(ys))
        #plt.show()

    def predict(self, dataset=None):
        """ return dbscan clustering of found local optimum

        :param dataset: dataset to be clustered in case fit was not called jetzt
        :return: clustering of the dataset
        """
        if self.params is None:
            self.fit(dataset)
        db = DBSCAN(eps=self.params[-1][0], min_samples=self.params[-1][1]).fit(self.dataset)
        return db.labels_

    def plot_result(self):
        """ plot scatter plot of found local optimum
        """
        labels = self.predict()
        nr_label = len(set(labels))
        hsv_tuples = [(x * 1.0 / nr_label, 0.9, 0.5) for x in range(nr_label)]
        rgb_tuples = list(map(lambda x: colorsys.hsv_to_rgb(*x), hsv_tuples))

        for label in set(labels):
            filtered_dataset = self.dataset[np.where(labels == label), ][0]

            if label == -1:
                # plot noise
                plt.scatter(filtered_dataset[:, 0], filtered_dataset[:, 1], c='black', s=50, marker='x')
                test =0
            else:
                # plot points in a cluster
                plt.scatter(filtered_dataset[:, 0], filtered_dataset[:, 1], c=rgb_tuples[label], s=50)


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    np.random.seed(17041991)

    #dataset, labels_true = generate_dataset("blobs", {"n_samples":100, "n_features":1000, "random_state":None})
    #dataset, labels_true = generate_dataset("moons", {"n_samples":100, "noise": 0.075, "random_state":None})
    #dataset, labels_true = generate_dataset("circles", {"n_samples":150, "noise": 0.01, "random_state":None})

    ## example datasets
    #dataset, labels = trex()
    #dataset, labels_true = aggregation()
    #dataset, labels_true = compound()
    #dataset, labels_true = r15()
    #dataset, labels_true = d31(size=1000)
    #dataset, labels_true = jain()
    #dataset, labels_true = pathbased()
    #dataset = t48k(size=4000)
    dataset, labels_true = spiral()
    #dataset, labels_true = flame()


    ## initialize alternating optimization
    ao = aoDBSCAN()


    fig = plt.figure(figsize=(8,15))



    #####################################################################################################
    #silhouette coefficient
    start_time = time.time()
    params = ao.fit_minPts(dataset, start_minPts=5, iterations=10, optimization="sil")
    end_time = time.time()
    print("sil:", end_time-start_time, "s")

    plt.subplot(3,2,1)
    ao.plot_learning()

    plt.subplot(3,2,2)
    ao.plot_result()

    labels_pred = ao.predict()
    homo, compl, vmeas = homogeneity_completeness_v_measure(labels_true, labels_pred)
    print(homo, compl, vmeas)


    #####################################################################################################
    #modified silhouette coefficient
    start_time = time.time()
    params = ao.fit_minPts(dataset, start_minPts=5, iterations=10, optimization="modsil")
    #params = ao.fit_minPts(dataset, start_minPts=1, iterations=10, optimization="sil")
    #params = ao.fit_minPts(dataset, start_minPts=1, iterations=10, optimization="edge")
    #params = ao.fit_eps(dataset, start_eps=1.4508618128546971, iterations=10, optimization="modsil")
    end_time = time.time()
    print("modsil:", end_time-start_time, "s")

    plt.subplot(3,2,3)
    ao.plot_learning()

    plt.subplot(3,2,4)
    ao.plot_result()

    labels_pred = ao.predict()
    homo, compl, vmeas = homogeneity_completeness_v_measure(labels_true, labels_pred)
    print(homo, compl, vmeas)


    #####################################################################################################
    #edge correlation
    start_time = time.time()
    params = ao.fit_minPts(dataset, start_minPts=5, iterations=10, optimization="edge")
    end_time = time.time()

    plt.subplot(3,2,5)
    ao.plot_learning()

    plt.subplot(3,2,6)
    ao.plot_result()

    print("edge:", end_time-start_time, "s")
    labels_pred = ao.predict()
    homo, compl, vmeas = homogeneity_completeness_v_measure(labels_true, labels_pred)
    print(homo, compl, vmeas)


    plt.show()