__author__ = 'Alex'

from sklearn.metrics import pairwise_distances
from hierarchicalDBSCAN.Plotting.VerticalSlider import VertSlider
from hierarchicalDBSCAN.Plotting.ColoredVerticalSlider import ColoredVertSlider
from hierarchicalDBSCAN.ClusteringAlgorithms.DendrogramTools import *
from hierarchicalDBSCAN.Datasets.DatasetLoader import *
from hierarchicalDBSCAN.Plotting.BasicPlotting import *
import random
import time
import networkx as nx
from hierarchicalDBSCAN.ClusteringAlgorithms.GraphTools import connected_component_subgraphs_core
import hierarchicalDBSCAN.ClusterValidation.InternalValidation as cv
from hierarchicalDBSCAN.ClusteringAlgorithms.epsDBScan import *

class epsGraphDBScan():
    """ Minimum Spanning Tree implementation of epsDBS
    """

    def __init__(self, eps):
        self.configure(eps)


    def configure(self, eps):
        """ configure the clustering object for the next run of fit()

        :param eps: eps value
        """
        self.eps = eps
        self.clustered_dataset = None
        self.clustergraph = None
        self.last_prediction_type = None
        self.simplified_clustergraph = None


    def get_clusterings(self):
        """

        :return:
        """
        levels = sorted(list(self.get_clusterlevels()), reverse=True)
        for level in levels:
            labels, clusterheights = self.predict_cut_height(level)
            yield level, labels, clusterheights


    def fit(self, dataset, dist_mat = None, min_minPts = None, bins = None):
        N = len(dataset)
        self.dataset = dataset

        # calculate distance matrix
        if dist_mat is None:
            distance_matrix = pairwise_distances(dataset)
        else:
            distance_matrix = dist_mat

        logging.info("calculate core distances")
        self.neighborhoods = {i: np.argwhere(distance_matrix[i,:]<=self.eps).flatten().tolist() for i in range(N)}
        self.core_set_size = {x: N - len(self.neighborhoods[x]) for x in self.neighborhoods}


        if min_minPts is None:
            min_minPts = 0
        min_minPts = N - min_minPts

        logging.info("build graph structure")
        graph = nx.empty_graph(N)

        for source in range(N):
             #add the whole neighboorhoodset with the weight
             if self.core_set_size[source] <= min_minPts:
                 for target in self.neighborhoods[source]:
                     if target > source:
                         dist = min(self.core_set_size[source], self.core_set_size[target])
                         graph.add_edge(source, target, weight=dist)

        logging.info("create spanning tree")
        T = nx.minimum_spanning_tree(graph)
        self.clustergraph = T
        self.clustergraph.graph['core_set_size'] = self.core_set_size


    def plot(self, cut_height):
        n = len(dataset)
        plt.subplot(2,1,1)
        position = {x: (self.dataset[x, 0], self.dataset[x, 1]) for x in range(n)}
        #nx.draw_networkx_edges(self.clustergraph, position)

        colors = [self.clustergraph.edge[u][v]["weight"] for (u, v) in self.clustergraph.edges()]
        nx.draw_networkx_edges(self.clustergraph, position, arrows=False, edge_color=colors, edge_cmap=plt.cm.cool, width=3)
        plt.gca().axis("equal")

        #plt.subplot(2,1,2)
        #labels, _ = self.predict_cut_height(cut_height)
        #plt.scatter(dataset[:,0], dataset[:,1], c=labels, s = 100)
        #plt.title("height {}".format(cut_height))
        #plt.gca().axis("equal")

        plt.show()

    def get_neighborhoodset(self, A,eps):
        """ return a list of points in radius of eps

        :param A: distance list
        :param eps: neighborhood radius
        :return: return index list of points in the neighborhood
        """
        return [n for n in range(len(A)) if A[n] <= eps]


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    #dataset, labels_true = generate_dataset("moons", {"n_samples": 1000, "noise": 0.075, "random_state":None})
    #dataset, labels_true = compound()
    #dataset, labels_true = aggregation()
    dataset = t48k()
    print("length dataset", len(dataset))

    db = epsGraphDBScan(eps=10)

    start_time = time.time()
    distance_matrix = pairwise_distances(dataset)
    end_time = time.time()
    print("distance matrix: ", end_time-start_time, "s")

    start_time = time.time()
    db.fit(dataset, distance_matrix)
    end_time = time.time()
    print("fitting minimum spanning tree: ", end_time-start_time, "s")


    db2 = epsHierarchicalDBScan(3)
    start_time = time.time()
    db.fit(dataset, distance_matrix)
    end_time = time.time()
    print("fitting cluster tree: ", end_time-start_time, "s")


