__author__ = 'Alex'

from matplotlib.widgets import Button
from sklearn.metrics import pairwise_distances
from hierarchicalDBSCAN.Plotting.VerticalSlider import VertSlider
from hierarchicalDBSCAN.Plotting.ColoredVerticalSlider import ColoredVertSlider
from hierarchicalDBSCAN.ClusteringAlgorithms.DendrogramTools import *
from hierarchicalDBSCAN.Datasets.DatasetLoader import *
from hierarchicalDBSCAN.Plotting.BasicPlotting import *


class hierarchicalDBScan():
    """ Base Class of hierarchical DBScan types minPtsHierarchicalDBScan and epsDBScan.

    Provides shared basic methods and plotting.
    """


    def simplify_clustergraph(self):
        """ removes successive nodes with same set of clustered "nodes"

        :return:
        """
        try:
            if self.clustergraph is None:
                raise ProcessingSequenceError('first run fit method')
        except ProcessingSequenceError:
            return
        self.simplified_clustergraph = simplify_clustergraph(self.clustergraph)



    def get_nr_of_clusterlevels(self):
        """
        :return: number of hierarchy levels
        """
        if self.simplified_clustergraph is None:
            self.simplify_clustergraph()
        return len(set([self.simplified_clustergraph.node[node]["height"] for node in self.simplified_clustergraph]))

    def get_clusterlevels(self):
        """
        :return: hierarchy levels
        """
        if self.simplified_clustergraph is None:
            self.simplify_clustergraph()
        return set([self.simplified_clustergraph.node[node]["height"] for node in self.simplified_clustergraph])


    def predict_cut_height(self, height, simplified=True):
        """
        :param method: choose method for cutting clustergraph
            "cut_height" : cuts the tree at given height (*args = )
        :param args: additional arguments for prediction
        :return: labellist
        """
        if simplified:
            return labels_for_cut_height(len(self.clustered_dataset), self.simplified_clustergraph, height)
        else:
            return labels_for_cut_height(len(self.clustered_dataset), self.clustergraph, height)


    def predict_cut_edge_quantile(self, clustergraph, height):
        """ Deprecated version of edge quantile cut. Use DifDenseHDBSCAN instead.
        """
        return cut_hierarchy_quantile(len(self.clustered_dataset), clustergraph, height)


    def plot(self, simplified=False, init_cut_height=None):
        """ Default plotting method
        """
        return self.plot_horizontal(simplified, init_cut_height)

    def plot_horizontal(self, simplified=True, init_cut_height=None):
        """ Method for drawing clustergraph output of advanced DB-Scan
        Plots the clustergraph, vertical slider and colored scatter plot.
        :param dataset: np-array containing the dataset
        :param clustergraph: graph containing all clusterlevels, e.g. a dendrogram

        :return: None
        """
        try:
            if self.clustergraph is None:
                raise ProcessingSequenceError('first run fit method')
        except ProcessingSequenceError:
            return

        if simplified and self.simplified_clustergraph is None:
            self.simplify_clustergraph()


        fig = plt.figure(figsize=(15,8))
        ax = plt.gca()

        plt.subplots_adjust(left=0.15, bottom=0.1)
        plt.subplot(1,2,1)
        if simplified:
            draw_advanced_dbscan(self.clustered_dataset, self.simplified_clustergraph)
        else:
            draw_advanced_dbscan(self.clustered_dataset, self.clustergraph)

        ylim = plt.gca().get_ylim()
        axcolor = 'lightgoldenrodyellow'
        ax_cutheight = plt.axes([0.05, 0.1, 0.02, 0.8], axisbg=axcolor)
        if init_cut_height is None:
            height_slider = VertSlider(ax_cutheight, '', ylim[0], ylim[1], valinit=ylim[0]+((ylim[1]-ylim[0])*0.5))
        else:
            print("ylim", ylim)
            init_cut_height = max(min(ylim[1], init_cut_height), ylim[0])
            print("init", init_cut_height)
            height_slider = VertSlider(ax_cutheight, '', ylim[0], ylim[1], valinit=init_cut_height)

        plt.subplot(1,2,1)
        l = plt.axhline(height_slider.val)  # l = horizontal line which has to be updated

        plt.subplot(1,2,2)
        if simplified:
            p = draw_clustered_scatter(self.clustered_dataset, self.simplified_clustergraph, 1) #scatter plot which has to be updated
        else:
            p = draw_clustered_scatter(self.clustered_dataset, self.clustergraph, 1) #scatter plot which has to be updated

        def update(val):
            # update horizontal line in tree plot
            l.set_ydata([val, val])

            #calculate cluster colors
            labels = self.predict_cut_height(val)
            self.labels = labels
            logging.info("final clusters: {0}".format(labels))

            n = len(set(labels))
            if -1 in labels:
                n -= 1

            hsv_tuples = [(x * 1.0 / n, 0.5, 0.5) for x in range(n)]
            rgb_tuples = list(map(lambda x: colorsys.hsv_to_rgb(*x), hsv_tuples))
            rgb_tuples.append((0.5,0.5,0.5))

            logging.debug("{0}".format(labels))
            #l != -1 => cluster, l = -2 => border of more than one cluster, l = -1 => not in any cluster
            color_list = [rgb_tuples[l] if l > -1 else (0.1,0.1,0.1) if l == -1 else (0.5,0.5,0.5) for l in labels]
            p.set_color(color_list)

            plt.gcf().canvas.draw_idle()

        height_slider.on_changed(update)
        update(height_slider.val)

        resetax = plt.axes([0.8, 0.025, 0.1, 0.04])
        button = Button(resetax, 'Reset', color=axcolor, hovercolor='0.975')

        def reset(event):
            height_slider.reset()
        button.on_clicked(reset)

        plt.show()


    def plot_hierarchy_cut(self, simplified=True, init_quantile=98.5):
        """ Method for drawing clustergraph output of advanced DB-Scan
        Plots the clustergraph, vertical slider and colored scatter plot.
        :param dataset: np-array containing the dataset
        :param clustergraph: graph containing all clusterlevels, e.g. a dendrogram

        :return: None
        """
        try:
            if self.clustergraph is None:
                raise ProcessingSequenceError('first run fit method')
        except ProcessingSequenceError:
            return



        fig = plt.figure(figsize=(15,8))
        ax = plt.gca()

        plt.subplots_adjust(left=0.15, bottom=0.1)
        plt.subplot(1,2,1)
        if simplified:
            self.simplify_clustergraph()
            ylim = draw_advanced_dbscan(self.clustered_dataset, self.simplified_clustergraph, color_edges=True)
            dendrogram_edges = [(u, v, self.clustergraph.node[u]["height"] - self.clustergraph.node[v]["height"])
                            for (u, v) in self.simplified_clustergraph.edges()]
        else:
            ylim = draw_advanced_dbscan(self.clustered_dataset, self.clustergraph, color_edges=True)
            dendrogram_edges = [(u, v, self.clustergraph.node[u]["height"] - self.clustergraph.node[v]["height"])
                            for (u, v) in self.clustergraph.edges()]

        #ylim = plt.gca().get_ylim()
        axcolor = 'lightgoldenrodyellow'
        ax_cutheight = plt.axes([0.05, 0.1, 0.1, 0.8], axisbg=axcolor)

        cut_height = np.percentile([h for _, _, h in dendrogram_edges], init_quantile)
        height_slider = ColoredVertSlider(ax_cutheight, '', ylim[0], ylim[1], valinit=cut_height)

        plt.subplot(1,2,2)
        if simplified:
            p = draw_clustered_scatter(self.clustered_dataset, self.simplified_clustergraph, 1) #scatter plot which has to be updated
        else:
            p = draw_clustered_scatter(self.clustered_dataset, self.clustergraph, 1) #scatter plot which has to be updated

        def update(val):
            #calculate cluster colors
            if simplified:
                labels = self.predict_cut_edge_quantile(self.simplified_clustergraph, val)
            else:
                labels = self.predict_cut_edge_quantile(self.clustergraph, val)

            logging.info("final clusters: {0}".format(labels))

            n = len(set([c for c in labels if c >= 0]))
            hsv_tuples = [(x * 1.0 / n, 0.9, 0.5) for x in range(n)]
            rgb_tuples = list(map(lambda x: colorsys.hsv_to_rgb(*x), hsv_tuples))

            mapping = {}
            map_idx = 0
            for label in set([c for c in labels if c >= 0]):
                mapping[label] = map_idx
                map_idx += 1

            color_list = [rgb_tuples[mapping[c]] if c >-1 else (0,0,0) if c == -1 else (0.5,0.5,0.5)  for c in labels]

            p.set_color(color_list)

            plt.gcf().canvas.draw_idle()

        height_slider.on_changed(update)
        update(height_slider.val)

        resetax = plt.axes([0.8, 0.025, 0.1, 0.04])
        button = Button(resetax, 'Reset', color=axcolor, hovercolor='0.975')

        def reset(event):
            height_slider.reset()
        button.on_clicked(reset)

        plt.show()