__author__ = 'Alex'

import time

from hierarchicalDBSCAN.ClusteringAlgorithms.hierarchicalDBScan import *


class minPtsHierarchicalDBScan(hierarchicalDBScan):
    """ HDBS implementation using clustertree based implementation
    """
    def __init__(self, minPts, max_radius = None):
        self.configure(minPts, max_radius)

    def configure(self, minPts, max_radius = None):
        """ configure Clustering object for the next run of fit()

        :param minPts: fix minPts
        :param max_radius: maximal eps to which the hierarchy should be created
        """
        self.minPts = minPts
        self.max_radius = max_radius
        self.clustered_dataset = None
        self.clustergraph = None
        self.last_prediction_type = None
        self.simplified_clustergraph = None


    def get_clusterings(self):
        """ returns all clustering levels of the simplified hierarchy

        :return: generator object for lust of clusterings
        """
        return labels_for_cut_height_generator(len(self.clustered_dataset), self.simplified_clustergraph, report_heights=True)


    def fit(self, dataset, bins = None, dist_mat = None):
        """ create clustergraph of dataset

        :param dataset: np array of points
        :param bins: number of bins for binning of the distance matrix
        :param dist_mat: precomputed distance matrix of the dataset
        """
        # reset variables
        self.simplified_clustergraph = None
        self.clustered_dataset = dataset
        UNIQUE_NODE_NAME = 0

        # calculate distance matrix or use precomputed version
        if dist_mat is None:
            distance_matrix = pairwise_distances(dataset)
        else:
            distance_matrix = dist_mat

        # allow for binning of the distances in the distance matrix
        # this reduces the number of hierarchy levels and can improve the run-time of the algorithm
        # however crucial levels cann be removed
        if bins is not None:
            bins = np.linspace(np.min(distance_matrix), np.max(distance_matrix), bins)
            distance_matrix = np.matrix([bins[x-1] for x in np.digitize(distance_matrix.flatten(), bins).reshape(distance_matrix.shape)])

        N = len(dataset)

        if self.max_radius is None:
            self.max_radius = np.inf

        # calculate distance triples
        sorted_distance_triples = list(np.ndenumerate(distance_matrix))
        sorted_distance_triples = sorted(sorted_distance_triples, key=itemgetter(1))
        nodes_in_range = {n: [] for n in range(N)}       # stores nodes in range till node becomes core

        core_points = set()
        self.clustergraph = nx.DiGraph()
        current_clusters = []
        old_height = -1

        # traverse through all distance triples and add links to clustergraph
        for (source, target), current_height in sorted_distance_triples:
            if current_height > self.max_radius:
                return
            if source == target:
                continue

            new_core = False
            if source not in core_points:
                nodes_in_range[source].append(target)
                if len(nodes_in_range[source]) == self.minPts:
                    # if core condition is reached add node to core list and process links
                    core_points.add(source)
                    new_core = True
                else:
                    # link will be processed when core condition is reached
                    continue

            # process new core
            if new_core:
                source_cluster_ids = [c for c in range(len(current_clusters)) if source in current_clusters[c]["nodes"]]

                #three possibilities:
                #new core is part of zero cluster
                # create new cluster with core
                if not source_cluster_ids:
                    new_node = {"nodes": set(nodes_in_range[source]+[source]),
                                "cores": set(nodes_in_range[source]+[source]).intersection(core_points),
                                "name": UNIQUE_NODE_NAME,
                                "height": current_height}
                    UNIQUE_NODE_NAME += 1

                    self.clustergraph.add_node(new_node["name"], new_node)
                    current_clusters.append(new_node)

                # new core is part of one cluster
                #  update cluster
                elif len(source_cluster_ids) == 1:
                    # nodes at same height = just update node
                    if current_height == current_clusters[source_cluster_ids[0]]["height"]:
                        node = current_clusters[source_cluster_ids[0]]
                        node["nodes"].update(nodes_in_range[source]+[source])
                        node["cores"].update([source])
                        self.clustergraph.node[node["name"]]["nodes"] = node["nodes"]
                        self.clustergraph.node[node["name"]]["cores"] = node["cores"]

                    # nodes at different height = create new node
                    else:
                        old_cluster = current_clusters[source_cluster_ids[0]]
                        new_node = {}
                        new_node["nodes"] = old_cluster["nodes"].union(set(nodes_in_range[source]+[source]))
                        new_node["cores"] = old_cluster["cores"].union(set([source]))
                        new_node["name"] = UNIQUE_NODE_NAME
                        new_node["height"] = current_height
                        UNIQUE_NODE_NAME += 1

                        self.clustergraph.add_node(new_node["name"], new_node)  # add node with attributes
                        self.clustergraph.add_edge(new_node["name"], current_clusters[source_cluster_ids[0]]["name"])

                        current_clusters.pop(source_cluster_ids[0])
                        current_clusters.append(new_node)

                # new core is part of more than one cluster
                #  merge clusters
                else:
                    new_node = {"nodes": set(nodes_in_range[source]+[source]),
                                "cores": set(nodes_in_range[source]+[source]).intersection(core_points),
                                "name": UNIQUE_NODE_NAME,
                                "height": current_height}
                    UNIQUE_NODE_NAME += 1
                    self.clustergraph.add_node(new_node["name"], new_node)

                    for c in sorted(source_cluster_ids, reverse=True):
                        # add edges to all children, remove old node
                        if current_height == current_clusters[c]["height"]:
                            for child in self.clustergraph.neighbors(current_clusters[c]["name"]):
                                self.clustergraph.add_edge(new_node["name"], child)
                            self.clustergraph.remove_node(current_clusters[c]["name"])

                        # add edge
                        else:
                            self.clustergraph.add_edge(new_node["name"], current_clusters[c]["name"])

                        new_node["nodes"].update(current_clusters[c]["nodes"])
                        new_node["cores"].update(current_clusters[c]["cores"])
                        current_clusters.pop(c)
                    current_clusters.append(new_node)

                    # set attributes of new_node
                    self.clustergraph.node[new_node["name"]]["height"] = new_node["height"]
                    self.clustergraph.node[new_node["name"]]["cores"] = new_node["cores"]
                    self.clustergraph.node[new_node["name"]]["nodes"] = new_node["nodes"]
                    self.clustergraph.node[new_node["name"]]["name"] = new_node["name"]
                continue

            # else process new link (would already have been processed if core was added)
            source_cluster_ids = [c for c in range(len(current_clusters)) if source in current_clusters[c]["nodes"]]
            target_cluster_ids = [c for c in range(len(current_clusters)) if target in current_clusters[c]["cores"]]

            # continue if target is already contained in the cluster
            if target in current_clusters[source_cluster_ids[0]]["nodes"]:
                continue

            #link from core to node = add node
            if not target_cluster_ids:
                # target has to be added to the cluster the source is in
                #  if height is the same just update the node
                if current_height == current_clusters[source_cluster_ids[0]]:
                    node = current_clusters[source_cluster_ids[0]]
                    node["nodes"].update([target])
                    self.clustergraph.node[node["name"]]["nodes"] = node["nodes"]

                # else add a new node is with same properties and new height
                else:
                    old_node = current_clusters[source_cluster_ids[0]]
                    new_node = {}
                    new_node["nodes"] = old_node["nodes"].copy().union([target])
                    new_node["cores"] = old_node["cores"].copy()
                    new_node["name"] = UNIQUE_NODE_NAME
                    new_node["height"] = current_height
                    UNIQUE_NODE_NAME += 1

                    self.clustergraph.add_node(new_node["name"], new_node)  # add node with attributes
                    self.clustergraph.add_edge(new_node["name"], current_clusters[source_cluster_ids[0]]["name"])

                    current_clusters.pop(source_cluster_ids[0])
                    current_clusters.append(new_node)

            # link from core to core = merge cluster source_cluster_ids[0] and target_cluster_ids[0]
            else:
                new_node = {"nodes": set(), "cores": set(), "name": UNIQUE_NODE_NAME, "height": current_height}
                UNIQUE_NODE_NAME += 1

                for c in sorted([source_cluster_ids[0], target_cluster_ids[0]], reverse=True):
                    # add edges to all children, remove old node
                    if current_height == current_clusters[c]["height"]:
                        for child in self.clustergraph.neighbors(current_clusters[c]["name"]):
                            self.clustergraph.add_edge(new_node["name"], child)
                        self.clustergraph.remove_node(current_clusters[c]["name"])

                    # add edge
                    else:
                        self.clustergraph.add_edge(new_node["name"], current_clusters[c]["name"])

                    new_node["nodes"].update(current_clusters[c]["nodes"])
                    new_node["cores"].update(current_clusters[c]["cores"])
                    current_clusters.pop(c)
                current_clusters.append(new_node)

                # set attributes of new_node
                self.clustergraph.node[new_node["name"]]["height"] = new_node["height"]
                self.clustergraph.node[new_node["name"]]["cores"] = new_node["cores"]
                self.clustergraph.node[new_node["name"]]["nodes"] = new_node["nodes"]
                self.clustergraph.node[new_node["name"]]["name"] = new_node["name"]


if __name__ == "__main__":
    #logging.basicConfig(level=logging.INFO)

    np.random.seed(17041991)

    #dataset, labels = aggregation()              #hierarchy cut init_quantil 99, horizontal_cut 2.67 (edge correlation), horizontal_cut 1.63 (density relation)
    #dataset, labels = d31()
    #dataset, labels = r15()
    #dataset, labels = compound()                #hierarchy cut init_quantil 99
    #dataset, labels = jain()                    #hierarchy cut init_quantil 99
    dataset, labels = spiral()                  #hierarchy cut init_quantil 99, horizontal cut 3.3 (handmade)
    #dataset = t48k(size=4000)
    #dataset, labels = pathbased()               #always shitty
    #dataset, labels_true = standard()               #always shitty
    #dataset, labels = flame()                   #always shitty
    #print("testset", len(dataset))
    #print("testset",dataset)


    db = minPtsHierarchicalDBScan(10)
    start_time = time.time()
    db.fit(dataset)
    end_time = time.time()
    print("binned: ", end_time-start_time)

    db.simplify_clustergraph()
    #print(sorted(list(db.get_clusterlevels()), reverse=True))
    db.plot_horizontal()
    db.plot_hierarchy_cut(simplified=False, init_quantile=98)