__author__ = 'Ajean: http://stackoverflow.com/questions/25934279/add-a-vertical-slider-with-matplotlib'


from matplotlib.widgets import AxesWidget

class VertSlider(AxesWidget):
    """
    A slider representing a floating point range

    The following attributes are defined
      *ax*        : the slider :class:`matplotlib.axes.Axes` instance

      *val*       : the current slider value

      *vline*     : a :class:`matplotlib.lines.Line2D` instance
                     representing the initial value of the slider

      *poly*      : A :class:`matplotlib.patches.Polygon` instance
                     which is the slider knob

      *valfmt*    : the format string for formatting the slider text

      *label*     : a :class:`matplotlib.text.Text` instance
                     for the slider label

      *closedmin* : whether the slider is closed on the minimum

      *closedmax* : whether the slider is closed on the maximum

      *slidermin* : another slider - if not *None*, this slider must be
                     greater than *slidermin*

      *slidermax* : another slider - if not *None*, this slider must be
                     less than *slidermax*

      *dragging*  : allow for mouse dragging on slider

    Call :meth:`on_changed` to connect to the slider event
    """
    def __init__(self, ax, label, valmin, valmax, valinit=0.5, valfmt='%1.2f',
                 closedmin=True, closedmax=True, slidermin=None,
                 slidermax=None, dragging=True, rounding=False, reverse=False,
                 **kwargs):
        """
        Create a slider from *valmin* to *valmax* in axes *ax*

        *valinit*
            The slider initial position

        *label*
            The slider label

        *valfmt*
            Used to format the slider value

        *closedmin* and *closedmax*
            Indicate whether the slider interval is closed

        *slidermin* and *slidermax*
            Used to constrain the value of this slider to the values
            of other sliders.

        additional kwargs are passed on to ``self.poly`` which is the
        :class:`matplotlib.patches.Rectangle` which draws the slider
        knob.  See the :class:`matplotlib.patches.Rectangle` documentation
        valid property names (e.g., *facecolor*, *edgecolor*, *alpha*, ...)
        """
        AxesWidget.__init__(self, ax)

        if reverse == True:
            self.reverse = True
            self.valmin = valmax
            self.valmax = valmin
        else:
            self.reverse = False
            self.valmin = valmin
            self.valmax = valmax

        if rounding==True:
            self.rounding = True
            self.val = round(valinit, 0)
            self.valinit = round(valinit,0)
            self.valfmt = '%1.0f'
        else:
            self.rounding = False
            self.val = valinit
            self.valinit = valinit
            self.valfmt = valfmt

        if reverse:
            self.poly = ax.axhspan(valmax, self.val, 0, 1, **kwargs)
        else:
            self.poly = ax.axhspan(valmin, valinit, 0, 1, **kwargs)

        self.vline = ax.axhline(self.valinit, 0, 1, color='r', lw=1)

        self.ax = ax
        ax.set_xticks([])
        ax.set_ylim((self.valmin, self.valmax))
        ax.set_yticks([])
        ax.set_navigate(False)

        self.connect_event('button_press_event', self._update)
        self.connect_event('button_release_event', self._update)
        if dragging:
            self.connect_event('motion_notify_event', self._update)
        self.label = ax.text(0.5, 1.03, label, transform=ax.transAxes,
                             verticalalignment='center',
                             horizontalalignment='center')

        self.valtext = ax.text(0.5, -0.03, valfmt % self.valinit,
                               transform=ax.transAxes,
                               verticalalignment='center',
                               horizontalalignment='center')

        self.cnt = 0
        self.observers = {}

        self.closedmin = closedmin
        self.closedmax = closedmax
        if reverse:
            self.slidermin = slidermax
            self.slidermax = slidermin
        else:
            self.slidermin = slidermin
            self.slidermax = slidermax
        self.drag_active = False

    def _update(self, event):
        """update the slider position"""

        if self.ignore(event):
            return

        if event.button != 1:
            return

        if event.name == 'button_press_event' and event.inaxes == self.ax:
            self.drag_active = True
            event.canvas.grab_mouse(self.ax)

        if not self.drag_active:
            return

        elif ((event.name == 'button_release_event') or
              (event.name == 'button_press_event' and
               event.inaxes != self.ax)):
            self.drag_active = False
            event.canvas.release_mouse(self.ax)
            return

        val = event.ydata
        if self.rounding:
            val = round(val, 0)

        if self.reverse:
            if val >= self.valmin:
                if not self.closedmin:
                    return
                val = self.valmin
            elif val <= self.valmax:
                if not self.closedmax:
                    return
        else:
            if val <= self.valmin:
                if not self.closedmin:
                    return
                val = self.valmin
            elif val >= self.valmax:
                if not self.closedmax:
                    return
                val = self.valmax

        if self.slidermin is not None and val <= self.slidermin.val:
            if not self.closedmin:
                return
            val = self.slidermin.val

        if self.slidermax is not None and val >= self.slidermax.val:
            if not self.closedmax:
                return
            val = self.slidermax.val

        self.set_val(val)

    def set_val(self, val):

        xy = self.poly.xy

        if self.reverse:
            xy[0] = 0, self.valmin
            xy[1] = 0, val
            xy[2] = 1, val
            xy[3] = 1, self.valmin
            xy[4] = 0, self.valmin
            self.poly.xy = xy
        else:
            xy[0] = 0, self.valmin
            xy[1] = 0, val
            xy[2] = 1, val
            xy[3] = 1, self.valmin
            xy[4] = 0, self.valmin
            self.poly.xy = xy

        self.valtext.set_text(self.valfmt % val)
        if self.drawon:
            self.ax.figure.canvas.draw()
        self.val = val
        #if not self.eventson:
        #    return
        for cid, func in self.observers.items():
            func(val)


    def on_changed(self, func):
        """
        When the slider value is changed, call *func* with the new
        slider position

        A connection id is returned which can be used to disconnect
        """
        cid = self.cnt
        self.observers[cid] = func
        self.cnt += 1
        return cid

    def disconnect(self, cid):
        """remove the observer with connection id *cid*"""
        try:
            del self.observers[cid]
        except KeyError:
            pass

    def reset(self):
        """reset the slider to the initial value if needed"""
        if (self.val != self.valinit):
            self.set_val(self.valinit)